#include "stdafx.h"

extern "C" {
OSVERSIONINFO OsVerInfo;
ULONG MajorVersion;
ULONG MinorVersion;
};

ULONG (__stdcall *NtCreateSymbolicLinkObject)(
                PHANDLE SymbolicLinkHandle,
                ACCESS_MASK DesiredAccess,
                POBJECT_ATTRIBUTES ObjectAttributes,
                PUNICODE_STRING DeviceName);

ULONG (__stdcall *NtOpenSymbolicLinkObject)(
                PHANDLE LinkHandle,
                ACCESS_MASK DesiredAccess,
                POBJECT_ATTRIBUTES ObjectAttributes);

ULONG (__stdcall *NtQuerySymbolicLinkObject)(
                HANDLE LinkHandle,
                PUNICODE_STRING LinkTarget,
                PULONG ReturnedLength);

ULONG (__stdcall *NtClose)(HANDLE Handle);

ULONG (__stdcall *NtOpenDirectoryObject)(
    OUT PHANDLE hDirectory,
	IN ACCESS_MASK AccessMask,
	IN POBJECT_ATTRIBUTES ObjectAttributes);


//--------------------------------------------------------------------
//
// LocateNDLLCalls
//
// Loads function entry points in NTDLL
//
//--------------------------------------------------------------------
extern "C"
BOOLEAN
LocateNTDLLCalls()
{

    if( !(NtCreateSymbolicLinkObject =
            (unsigned long (__stdcall *)(void ** ,unsigned long,struct _OBJECT_ATTRIBUTES *,struct _UNICODE_STRING *))
            GetProcAddress( GetModuleHandle("ntdll.dll"),
            "NtCreateSymbolicLinkObject" )) ) {
        return FALSE;
    }
    if( !(NtOpenSymbolicLinkObject =
            (unsigned long (__stdcall *)(void ** ,unsigned long,struct _OBJECT_ATTRIBUTES *))
            GetProcAddress( GetModuleHandle("ntdll.dll"),
            "NtOpenSymbolicLinkObject" )) ) {
        return FALSE;
    }
    if( !(NtQuerySymbolicLinkObject =
            (unsigned long (__stdcall *)(void *,struct _UNICODE_STRING *,unsigned long *))
            GetProcAddress( GetModuleHandle("ntdll.dll"),
            "NtQuerySymbolicLinkObject" )) ) {
        return FALSE;
    }
    if( !(NtClose =
            (unsigned long (__stdcall *)(void *))
            GetProcAddress( GetModuleHandle("ntdll.dll"),
            "NtClose" )) ) {
        return FALSE;
    }
    if( !(NtOpenDirectoryObject =
            (unsigned long (__stdcall *)(void ** ,unsigned long,struct _OBJECT_ATTRIBUTES *))
            GetProcAddress( GetModuleHandle("ntdll.dll"),
            "NtOpenDirectoryObject" )) ) {
        return FALSE;
    }

    return TRUE;
} // end LocateNTDLLCalls()

extern "C"
BOOLEAN
ParseSymLink1(
    IN PUNICODE_STRING _SrcUniName,
    OUT PUNICODE_STRING DstPathName
    )
{
    BOOLEAN  IsSuccess = FALSE;
    NTSTATUS NtStatus  = 0;
    __try
    {
        HANDLE         SymLinkHandle      = NULL;
        HANDLE         ObDirHandle = NULL;

        PWCHAR         SymLinkNameBuffer = NULL; 
        PWCHAR         NewNameBuffer = NULL; 
        UNICODE_STRING SrcUniName = *_SrcUniName;
        UNICODE_STRING DstUniName;
        UNICODE_STRING PathName;

        ULONG i;
        ULONG prev_i=0;
        ULONG lim;

        OBJECT_ATTRIBUTES   ObAttributes;

        SymLinkNameBuffer = (PWCHAR)ExAllocatePool(NonPagedPool, MAX_UNI_PATH + 2);
        NewNameBuffer     = (PWCHAR)ExAllocatePool(NonPagedPool, MAX_UNI_PATH + 2);
        PathName.Buffer   = NewNameBuffer;

        memcpy(SymLinkNameBuffer, SrcUniName.Buffer, SrcUniName.MaximumLength);
        memcpy(NewNameBuffer, SrcUniName.Buffer, SrcUniName.MaximumLength);

        if(SymLinkNameBuffer && NewNameBuffer) {
            lim = SrcUniName.Length / sizeof(WCHAR);
        }

        while (SymLinkNameBuffer && NewNameBuffer)
        {
            for(i=prev_i+1; i<lim; i++) {
                if(NewNameBuffer[i] == '\\') {
                    break;
                }
            }
            // Open the Directory:
            memcpy(SymLinkNameBuffer, NewNameBuffer, (prev_i ? prev_i : 1)*sizeof(WCHAR));
            SymLinkNameBuffer[prev_i ? prev_i : 1] = 0;
            RtlInitUnicodeString(&SrcUniName, SymLinkNameBuffer);

            InitializeObjectAttributes(&ObAttributes,
                                       &SrcUniName,
                                       OBJ_CASE_INSENSITIVE,
                                       (HANDLE)NULL,
                                       (PSECURITY_DESCRIPTOR)NULL);

            NtStatus = NtOpenDirectoryObject(&ObDirHandle,
                                             DIRECTORY_QUERY,
                                             &ObAttributes);

            if (NT_SUCCESS(NtStatus))
            {
                memcpy(SymLinkNameBuffer, NewNameBuffer+prev_i+1, (i-prev_i-1)*sizeof(WCHAR));
                SymLinkNameBuffer[i-prev_i-1] = 0;
                RtlInitUnicodeString(&SrcUniName, SymLinkNameBuffer);

                InitializeObjectAttributes(&ObAttributes,
                                           &SrcUniName,
                                           OBJ_CASE_INSENSITIVE,
                                           (HANDLE)ObDirHandle,
                                           (PSECURITY_DESCRIPTOR)NULL);

                NtStatus = NtOpenSymbolicLinkObject(&SymLinkHandle,
                                                    SYMBOLIC_LINK_QUERY,
                                                    &ObAttributes);
                if (NT_SUCCESS(NtStatus))
                {
                    memset(SymLinkNameBuffer, 0, MAX_UNI_PATH);
                    DstUniName.Length        = 0;
                    DstUniName.MaximumLength = MAX_UNI_PATH;
                    DstUniName.Buffer        = SymLinkNameBuffer;

                    NtStatus = NtQuerySymbolicLinkObject(SymLinkHandle,
                                                         &DstUniName,
                                                         NULL);
                    // Close handle in current process context:
                    NtClose((HANDLE) SymLinkHandle);
                } else {
                    // next pattern
                    prev_i = i;
                    continue;
                }

                if (NT_SUCCESS(NtStatus))
                {
                    if(DstUniName.Length &&
                       SymLinkNameBuffer[(DstUniName.Length-1)/sizeof(WCHAR)] == '\\') {
                        DstUniName.Length--;
                        SymLinkNameBuffer[(DstUniName.Length)/sizeof(WCHAR)] = 0;
                    }
                    memcpy(SymLinkNameBuffer+(DstUniName.Length/sizeof(WCHAR)),
                           NewNameBuffer+i, (lim-i)*sizeof(WCHAR));
                    DstUniName.Length += (USHORT)((lim-i)*sizeof(WCHAR));

                    memcpy(NewNameBuffer, SymLinkNameBuffer, DstUniName.Length);
                    NewNameBuffer[DstUniName.Length/sizeof(WCHAR)] = 0;
                    PathName.Length = DstUniName.Length;

                    prev_i = 0;
                    i = 0;
                    lim = PathName.Length/sizeof(WCHAR);
                    
                } else {
                    // Quit
                    NewNameBuffer = NULL;
                }

                // Close handle in current process context:
                NtClose((HANDLE) ObDirHandle);
            } else {
                break;
            }
            if(i>=lim) {
                break;
            }
            break;
        }
        RtlInitUnicodeString(DstPathName, NewNameBuffer);
    }
    __except(EXCEPTION_EXECUTE_HANDLER)
    {
        IsSuccess = FALSE;
    }

    return IsSuccess;
} // end ParseSymLink1()

